// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors
 */
// </header>
package fr.unicaen.openmap.app;

import java.io.IOException;
import java.net.MalformedURLException;

import java.util.logging.Level;
import java.util.logging.Logger;


import com.bbn.openmap.Layer;
import com.bbn.openmap.LayerHandler;

import com.bbn.openmap.gui.OpenMapFrame;
import com.bbn.openmap.layer.OMGraphicHandlerLayer;
import com.bbn.openmap.omGraphics.OMGraphicHandler;
import com.bbn.openmap.proj.Proj;
import com.bbn.openmap.proj.Projection;
import com.bbn.openmap.proj.ProjectionFactory;
import com.bbn.openmap.util.ArgParser;
import com.bbn.openmap.util.Debug;

import fr.unicaen.openmap.PropertyHandler;
import fr.unicaen.openmap.gui.ThemaMapPanel;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer;
import fr.unicaen.openmap.layer.utils.LayerLoader;
import fr.unicaen.openmap.util.HtmlGenerator;
import fr.unicaen.openmap.util.ImageExporter;
import fr.unicaen.openmap.util.TooManyLegendException;
import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Properties;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JProgressBar;

/**
 * The OpenMap application framework. This class creates a PropertyHandler that
 * searches the classpath, config directory and user's home directory for an
 * openmap.properties file, and creates the application based on the contents of
 * the properties files. It also creates an MapPanel and an OpenMapFrame to be
 * used for the application and adds them to the MapHandler contained in the
 * MapPanel. All other components are added to that MapHandler as well, and they
 * use the MapHandler to locate, connect and communicate with each other.
 */
public class ThemaMapBenchTest extends ThemaMap {

    static JDialog progressDialog;
    static JProgressBar progressBar;
    static ArrayList<String> options;
    static final String propertiesOption = "-properties";
    static final String folderOption = "-folder";
    static final String htmlOption = "-html";
    static final String formatOption = "-format";
    static final String delayOption = "-delay";
    static HashMap<String, String> map;
    static PropertyHandler propertyHandler;
    static {
        options = new ArrayList<String>();
        options.add(propertiesOption);
        options.add(folderOption);
        options.add(htmlOption);
        options.add(formatOption);
        options.add(delayOption);
    }

    /**
     * Create a new OpenMap framework object - creates a MapPanel, OpenMapFrame,
     * and brings up the layer palettes that are being told to be open at
     * startup. The MapPanel will create a PropertiesHandler that will search
     * for an openmap.properties file.
     */
    public ThemaMapBenchTest() {
        this(null);
    }

    /**
     * Create a new OpenMap framework object - creates a MapPanel, OpenMapFrame,
     * and brings up the layer palettes that are being told to be open at
     * startup. The properties in the PropertyHandler will be used to configure
     * the application. PropertyHandler may be null.
     */
    public ThemaMapBenchTest(PropertyHandler propertyHandler) {
        super(propertyHandler);
    }

    /**
     * Create and return an OpenMap object that uses a standard PropertyHandler
     * to configure itself. The OpenMap object has a MapHandler that you can use
     * to gain access to all the components.
     *
     * @return OpenMap
     * @see #getMapHandler
     */
    public static ThemaMapBenchTest create() {
        return new ThemaMapBenchTest(null);
    }

    /**
     * Create and return an OpenMap object that uses a standard PropertyHandler
     * to configure itself. The OpenMap object has a MapHandler that you can use
     * to gain access to all the components.
     *
     * @return OpenMap
     * @see #getMapHandler
     */
    public static ThemaMapBenchTest create(String propertiesFile) {
        Debug.init();
        //System.out.println(propertiesFile);
        propertyHandler = null;

        if (propertiesFile != null) {
            try {
                //System.out.println("THEMAMAPBENCHTEST create with propertyFile = " + propertiesFile);
                propertyHandler = new PropertyHandler(propertiesFile);
            } catch (MalformedURLException murle) {
                Debug.error(murle.getMessage());
                murle.printStackTrace();
                propertyHandler = null;
            } catch (IOException ioe) {
                Debug.error(ioe.getMessage());
                ioe.printStackTrace();
                propertyHandler = null;
            }
        }

        return new ThemaMapBenchTest(propertyHandler);
    }

    @Override
    protected void showInFrame() {
        ((ThemaMapPanel) mapPanel).create();
        OpenMapFrame omf = new OpenMapFrame(true);
        //omf.setExtendedState(JFrame.MAXIMIZED_BOTH);
        setWindowListenerOnFrame(omf);
        getMapHandler().add(omf);
        //SVGExporter.initMapHandler(getMapHandler());//must be done here?
        omf.setVisible(true);
        mapPanel.getMapBean().showLayerPalettes();
        Debug.message("basic", "OpenMap: READY");
        omf.setSize(omf.getToolkit().getScreenSize());
        progressDialog = new JDialog(omf, true);
        progressDialog.setSize(300, 100);
        progressBar = new JProgressBar(0, 100);
        JPanel pProgress = new JPanel();
        pProgress.add(progressBar);
        progressDialog.add(pProgress);
        progressDialog.setTitle("Export processing ...");
        progressBar.setIndeterminate(true);
        progressDialog.setVisible(true);
        // omf.setVisible(false);
    }

    public static HashMap<String, String> parseArgs(String[] args) {
        int a = 0;
//        System.out.println("-----args----");
//        for(String s: args){
//            System.out.println(a+"->"+s);
//            a++;
//        }
//        System.out.println("-------------");
        HashMap<String, String> map = new HashMap<String, String>();
        for (String option : options) {
            int index = getOptionIndex(option, args);
            if (index != -1) {
                String value = args[index + 1];
                if(!options.contains(value))
                    map.put(option, value);
            }
        }
        return map;
    }

    public static int getOptionIndex(String option, String[] args) {
        int i = 0;
        for (String a : args) {
            if (a.equals(option)) {
                return i;
            }
            i++;
        }
        return -1;
    }

    /**
     * The main OpenMap application.
     */
    /*public static void main(String args[]){
    ThemaMap th = ThemaMap.create("exemple.properties");
    }*/
    static public void main(String args[]) {
        try {
            map = ThemaMapBenchTest.parseArgs(args);
            ThemaMapBenchTest.create(map.get(propertiesOption));
            ImageExporter.ImageType imageType;
            if(map.get(formatOption) != null){
                if(map.get(formatOption).equals("svg"))
                    imageType = ImageExporter.ImageType.SVG;
                else
                   imageType = ImageExporter.ImageType.JPG;
            }
            else
                imageType = ImageExporter.ImageType.JPG; //default

            IsReadyTheadAndSubMachineGun irt =
                    new IsReadyTheadAndSubMachineGun(
                        map.get(folderOption),imageType, map.get(htmlOption));
            irt.start();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public static void ImageSubMachineGun(String folder, ImageExporter.ImageType type) {
        //System.out.println("-------------------------->IMAGE SUB MACHINE GUN");
        File file = new File(folder);
        if (file.isDirectory()) {
            File[] files = file.listFiles();
            for (File f : files) {
                if (!f.getName().endsWith(".properties")) {
                    continue;
                }
                ThemaMapBenchTest.closeCurrentProject();
                ThemaMapBenchTest.loadPropsFromFile(f);
                            try {
                Thread.sleep(Integer.parseInt(map.get(delayOption)));
            } catch (InterruptedException ex) {
                Logger.getLogger(ThemaMapBenchTest.class.getName()).log(Level.SEVERE, null, ex);
            }
                String imgFileName = f.getPath() + "." + type.toString().toLowerCase();
                ImageExporter.exportInFile(imgFileName, type);
            }
        }
    }

    public static void progressCompleted() {
        ThemaMapBenchTest.progressBar.setIndeterminate(false);
        ThemaMapBenchTest.progressBar.setValue(100);
        ThemaMapBenchTest.progressDialog.setTitle("SVG Export completed");
    }

    public static void closeCurrentProject() {
        try {
            LayerHandler layerHandler = (LayerHandler) getMapHandler().get("com.bbn.openmap.LayerHandler");
            for (Layer l : layerHandler.getLayers()) {
                //manager.getLayersPanel().deleteLayer(l);
                layerHandler.removeLayer(l);
            }
        } catch (NullPointerException exc) {//mapHandler not Ready
            closeCurrentProject();
        }
    }

    public static void loadPropsFromFile(File file) {

        try {
            //System.out.println("CHARGEMENT DU FICHIER DE PROPERTIES " + file.getName());
            LayerLoader layerLoader = new LayerLoader(file.toURI().toURL(), (LayerHandler) getMapHandler().get("com.bbn.openmap.LayerHandler"));
            String[] layersName = layerLoader.findAvailableLayers();
            ArrayList<String> rLayers = new ArrayList<String>(Arrays.asList(layersName));
            Collections.reverse(rLayers);
            //Collections.reverse(layers);
            ArrayList<Layer> layers = new ArrayList<Layer>();
            for (String selection : rLayers) {
                layers.add(layerLoader.initLayer(selection));

            }
            Properties newProps = PropertyHandler.fetchProperties(file.toURI().toURL());
            String scale = newProps.getProperty("openmap.Scale");
            Double latCenter = Double.parseDouble(newProps.getProperty("openmap.Latitude"));
            Double lonCenter = Double.parseDouble(newProps.getProperty("openmap.Longitude"));
            Projection projection = mapPanel.getMapBean().getProjection();
            Proj proj = (Proj) projection;
            proj.setScale(Float.parseFloat(scale));
            proj.setCenter(latCenter, lonCenter);
//            System.out.println("APPLICATION DE LA PROJECTION D'ECHELLE " + scale + " et de CENTRE " + latCenter+";"+lonCenter);
//            System.out.println("*************************************FIN CHARGEMENT**************************");
            mapPanel.getMapBean().repaint();
            for(Layer l : layers){
                l.setProjection(proj);
                if(l instanceof OMGraphicHandlerLayer)
                    ((OMGraphicHandlerLayer)l).doPrepare();
//                if(l instanceof ThematicDataDisplayLayer)
//                    ((ThematicDataDisplayLayer)l).refreshGraphics();
            }

            String test = newProps.getProperty("openmap." + LayerHandler.layersProperty);
            if (test == null) {
                throw new IOException("Doesn't seem like a valid properties file");
            }

            // Just reset the projection and layers, not the
            // components.
            // layerHandler.getPropertyHandler().loadProjectionAndLayers(mapHandler, newProps);


        } catch (FileNotFoundException fnfe) {
            fnfe.printStackTrace();
        } catch (MalformedURLException fnfe) {
            fnfe.printStackTrace();
        } catch (TooManyLegendException fnfe) {
            fnfe.printStackTrace();
        } catch (NullPointerException ne) {
            ne.printStackTrace();
        } catch (IOException ioe) {
            ioe.printStackTrace();
//            InformationDelegator id = (InformationDelegator) getMapHandler().get("com.bbn.openmap.InformationDelegator");
//            if (id != null) {
//                id.displayMessage("Error while loading file ...",
//                        "Error while loading file " + file.getAbsolutePath() + "\n" + ioe.getMessage());
        }
    }
}
/**
 * Tests if the current Themamap map handler is ready. The image export is performed when the maphandler is ready
 * @author jchauvea
 */
class IsReadyTheadAndSubMachineGun extends Thread {

    private boolean isReady = false;
    private String folder;
    private ImageExporter.ImageType type;
    private String htmlFile;

    public IsReadyTheadAndSubMachineGun(String folder, ImageExporter.ImageType type, String htmlFile) {
        this.folder = folder;
        this.type = type;
        this.htmlFile = htmlFile;
    }

    public void run() {
        while (!isReady) {
            if (!(ThemaMapBenchTest.getMapHandler() == null) && !(ThemaMapBenchTest.getMapHandler().get("com.bbn.openmap.LayerHandler") == null)) {
                isReady = true;
                ImageExporter.initMapHandler(ThemaMapBenchTest.getMapHandler());
                ThemaMapBenchTest.ImageSubMachineGun(folder, type);
                ThemaMapBenchTest.progressCompleted();
                if(htmlFile != null)
                    HtmlGenerator.generateHTML(folder, htmlFile, type);

            } else {
                try {
                    Thread.sleep(100);
                } catch (InterruptedException ex) {
                    Logger.getLogger(ThemaMapBenchTest.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
    }
}


    
