// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Set;
import java.util.HashSet;
import java.util.Collections;
import java.util.Comparator;
import java.util.regex.Pattern;

import javax.swing.JComponent;
import javax.swing.JPanel;

import com.bbn.openmap.util.Debug;
import com.bbn.openmap.omGraphics.OMPoly;
import com.bbn.openmap.omGraphics.OMLine;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.geo.Geo;

import com.bbn.openmap.util.PropUtils;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;
import fr.unicaen.openmap.dataStorage.gui.DataTableStoragePanel;
import fr.unicaen.openmap.dataStorage.BoundingBox;

import fr.unicaen.openmap.dataStorage.gui.ContoursPanel;
import fr.unicaen.openmap.dataStorage.triangulation.paulchew.DataPoint;

import fr.unicaen.openmap.jts.JTS;
import fr.unicaen.openmap.util.Util;
import java.util.Properties;

public class Contours extends InformationDataStorage implements RemovableDataStorage, DataStorageListener {

    public static final String dataStorageClassPropertyString = "dataStorageClass";
    public static final String contoursDSPropertyString = "contours";
    public static final String geometryColumnNamePropertyString = "geometryColumnName";
    public static final String dataColumnNamePropertyString = "dataColumnName";
    public static final String data0ColumnNamePropertyString = "data0ColumnName";
    public static final String data1ColumnNamePropertyString = "data1ColumnName";
    public static final String data2ColumnNamePropertyString = "data2ColumnName";
    public static final String contoursValuesPropertyString = "contourValues";
    public static final String IntervalSuffix = "_interv";
    public static final String withLinesPropertyString = "withlines";
    public static final String withPolysPropertyString = "withpolys";
    protected String isoPolyColumnNamePropertyString = "isopolygon";
    protected String contourColumnNamePropertyString = "isoline";

    protected DataStorage dataStorage;
    protected String geometryColumnName = "triangle";
    private String dataColumnName = "";
    protected String data0ColumnName = "_0";
    protected String data1ColumnName = "_1";
    protected String data2ColumnName = "_2";
    private String contourColumnName = "isoline";
    private String isoPolyColumnName = "isopolygon";
   
    protected ArrayList<Double> contoursValues;
    protected boolean gridIsTriangular = true;

    protected boolean addMesh = false;
    protected String meshColumnName = "mesh"; // mainly for debugging

    private boolean withPolygons = false;
    private boolean withLines = true;

    public Contours(DataStorageListener dsl) {
        super(dsl);
        dataTableModel = new DataTableModel(dataTable);
        dtsp = new DataTableStoragePanel(dataTableModel);
    }

    public Contours(DataStorageListener dsl, DataStorage dataStorage) {
        super(dsl);
        contoursValues = new ArrayList<Double>();
        if (contoursValues.isEmpty()) {
            contoursValues.add(4.0);
        }
        this.dataStorage = dataStorage;
        // we are now the listener
        dataStorage.dsl = this;
        dataTableModel = new DataTableModel(dataTable);
        dtsp = new DataTableStoragePanel(dataTableModel);
    }

    @Override
    public boolean isReady() {
        return dataStorage != null && dataStorage.isReady();
    }
    protected ArrayList<DataObject> dataObjects;

    public void initialize() {
        if (!isReady()) {
            System.err.println("[Contours] can't initialize because dataStorage is not ready !!!!");
            return;
        }
	if (gridIsTriangular)
	    initFromTriangularGrid();

        else {
	    //....
	    System.err.println("[Contours] ### regular grid not implemented yet");
	}
    }

    private void initFromTriangularGrid() {
	    dataObjects = new ArrayList<DataObject>();
	    ArrayList<DataObject> dataLineObjects = new ArrayList<DataObject>();

	    clearConvexHullTriangleSet();
	
	    Debug.message("DataStorage", "Contours "+ getDataColumnName() + " " +data0ColumnName+" "+data1ColumnName+ " "+ data2ColumnName + " values " + getContoursValues());
	    for (int i = 0; i < dataStorage.getCount(); i++) {
		DataObject data = dataStorage.getElementAt(i);
		OMDataLocation omdl = data.getOMDataLocation(geometryColumnName);
		if (omdl == null)
		    continue;
		double d0 = Util.parseDouble(data.getData(data0ColumnName));
		double d1 = Util.parseDouble(data.getData(data1ColumnName));
		double d2 = Util.parseDouble(data.getData(data2ColumnName));
		OMPoly triangle;
		try {
		    triangle = (OMPoly)omdl.getOMGraphic();
		} catch (ClassCastException e) {
		    System.err.println("### [Contours] bad triangle: " + omdl.getOMGraphic());
		    continue;
		}

                Object convexHullEdge = getConvexHullEdge(triangle);

                double[] rawllradianspts = triangle.getLatLonArray();
                double lat0 = Geo.degrees(rawllradianspts[0]);
                double lon0 = Geo.degrees(rawllradianspts[1]);
                double lat1 = Geo.degrees(rawllradianspts[2]);
                double lon1 = Geo.degrees(rawllradianspts[3]);
                double lat2 = Geo.degrees(rawllradianspts[4]);
                double lon2 = Geo.degrees(rawllradianspts[5]);

		for (double contourValue : contoursValues) {
		    int intersectionCase = computeIntersectionCase(d0, d1, d2, contourValue);
		    Debug.message("Contours2", "computeIntersectionCase("+d0+", "+d1+", "+d2+", "+contourValue+") => "+intersectionCase);
		    if (intersectionCase == 0)
			continue;
		    OMLine oml = intersection(
					      lat0, lon0, d0,
					      lat1, lon1, d1,
					      lat2, lon2, d2,
					      contourValue, intersectionCase, 
					      convexHullEdge);
		    if (oml == null) // this case should never happen...
			continue;

		    DataObject contourData = new DataObject();
		    contourData.addData(getDataColumnName(), contourValue);
		    OMDataLocation omdlContour = mkOMDataLocation(oml);
		    contourData.addOMDataLocation(getContourColumnName(), omdlContour);
		    if (addMesh)
			contourData.addOMDataLocation(meshColumnName, omdl);
 
		    dataLineObjects.add(contourData);
		    if (withLines)
			dataObjects.add(contourData);
		    Debug.message("Contours", "Add line for "+contourValue + " " + omLineToString(oml));
		}
	    }
	    addConvexHullLinesToDataObjects(dataLineObjects, withLines ? dataObjects : null);
	    
	    // 
	    if (withPolygons) {	
		String intervalDataColumnName =  getDataColumnName()+IntervalSuffix;
		ArrayList<DataObject> polys = JTS.polygonize(dataLineObjects, contourColumnName, getDataColumnName(), getIsoPolyColumnName(), intervalDataColumnName);
		if (polys != null) {
		    for (DataObject p : polys)
			p.addData(intervalDataColumnName, setToInterval(p.getData(intervalDataColumnName)));
		    dataObjects.addAll(polys);
		}
	    }

	    setTable();

    }

    private static Object setToInterval(Object s) {
	if (! (s instanceof Set))
	    return s;
	double min = Double.MAX_VALUE;
	double max = Double.MIN_VALUE;
	for(Object o : (Set)s) {
	    double d = Util.parseDouble(o);
	    if (d < min)
		min = d;
	    if (d > max)
		max = d;
	}
	ArrayList<Double> ret = new ArrayList<Double>();
	ret.add(min);
	ret.add(max);
	return ret;
    }

    private static OMDataLocation mkOMDataLocation(OMLine oml) {
        Centroid c = new Centroid((oml.getLL()[1] + oml.getLL()[3]) / 2, (oml.getLL()[0] + oml.getLL()[2]) / 2);
        BoundingBox bb = new BoundingBox(oml.getLL()[1], oml.getLL()[0], oml.getLL()[1], oml.getLL()[0]);
        bb.extendTo(oml.getLL()[3], oml.getLL()[2]);
        return new OMDataLocation(oml, c, bb);
    }

    private int computeIntersectionCase(double z0, double z1, double z2, double z) {

        double zMin = Math.min(Math.min(z0, z1), z2);
        double zMax = Math.max(Math.max(z0, z1), z2);

        if (z < zMin || zMax < z) {
            // no intersection
            return 0;
        }

        double zDist0 = z0 - z, zDist1 = z1 - z, zDist2 = z2 - z;

        int[][][] _intersectionCases = {{
                {0, 0, 8}, {0, 2, 5}, {7, 6, 9}
            }, {
                {0, 3, 4}, {1, 3, 1}, {4, 3, 0}
            }, {
                {9, 6, 7}, {5, 2, 0}, {8, 0, 0}
            }};

        return _intersectionCases[(int) (Math.signum(zDist0) + 1)][(int) (Math.signum(zDist1) + 1)][(int) (Math.signum(zDist2) + 1)];
    }

    private OMLine intersection(double lat0, double lon0, double z0,
            double lat1, double lon1, double z1,
            double lat2, double lon2, double z2,
            double z, int intersectionCase,
            Object convexHullEdge) {
        if (intersectionCase == 0) {
            return null;
        }
        Debug.message("Contours2", "intersection(" + z0 + " " + z1 + " " + z2 + " ? " + z + "), intersection case " + intersectionCase + " ([" + lat0 + ", " + lon0 + "] [" + lat1 + ", " + lon1 + "] [" + lat2 + ", " + lon2 + "]) " + convexHullEdge);

        double iLat0 = 0, iLon0 = 0, iLat1 = 0, iLon1 = 0;
        switch (intersectionCase) {
            case 1: // Line between vertices 1 and 2
                iLat0 = lat0;
                iLon0 = lon0;
                iLat1 = lat1;
                iLon1 = lon1;
                break;
            case 2: // Line between vertices 2 and 3
                iLat0 = lat1;
                iLon0 = lon1;
                iLat1 = lat2;
                iLon1 = lon2;
                break;
            case 3: // Line between vertices 3 and 1
                iLat0 = lat2;
                iLon0 = lon2;
                iLat1 = lat0;
                iLon1 = lon0;
                break;
            case 4: // Line between vertex 1 and side 2-3
                iLat0 = lat0;
                iLon0 = lon0;
                iLat1 = linearApproximation(z, lat1, lat2, z1, z2);
                iLon1 = linearApproximation(z, lon1, lon2, z1, z2);
                if (convexHullEdge != null) {
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat1, lon1, iLat1, iLon1, lat2, lon2);
                }
                break;
            case 5: // Line between vertex 2 and side 3-1
                iLat0 = lat1;
                iLon0 = lon1;
                iLat1 = linearApproximation(z, lat2, lat0, z2, z0);
                iLon1 = linearApproximation(z, lon2, lon0, z2, z0);
                if (convexHullEdge != null) {
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat2, lon2, iLat1, iLon1, lat0, lon0);
                }
                break;
            case 6: //  Line between vertex 3 and side 1-2
                iLat0 = lat2;
                iLon0 = lon2;
                iLat1 = linearApproximation(z, lat0, lat1, z0, z1);
                iLon1 = linearApproximation(z, lon0, lon1, z0, z1);
                if (convexHullEdge != null) {
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat0, lon0, iLat1, iLon1, lat1, lon1);
                }
                break;
            case 7: // Line between sides 1-2 and 2-3
                iLat0 = linearApproximation(z, lat0, lat1, z0, z1);
                iLon0 = linearApproximation(z, lon0, lon1, z0, z1);
                iLat1 = linearApproximation(z, lat1, lat2, z1, z2);
                iLon1 = linearApproximation(z, lon1, lon2, z1, z2);
                if (convexHullEdge != null) {
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat0, lon0, iLat0, iLon0, lat1, lon1);
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat1, lon1, iLat1, iLon1, lat2, lon2);
                }
                break;
            case 8: // Line between sides 2-3 and 3-1
                iLat0 = linearApproximation(z, lat1, lat2, z1, z2);
                iLon0 = linearApproximation(z, lon1, lon2, z1, z2);
                iLat1 = linearApproximation(z, lat2, lat0, z2, z0);
                iLon1 = linearApproximation(z, lon2, lon0, z2, z0);
                if (convexHullEdge != null) {
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat1, lon1, iLat0, iLon0, lat2, lon2);
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat2, lon2, iLat1, iLon1, lat0, lon0);
                }
                break;
            case 9: // Line between sides 3-1 and 1-2
                iLat0 = linearApproximation(z, lat2, lat0, z2, z0);
                iLon0 = linearApproximation(z, lon2, lon0, z2, z0);
                iLat1 = linearApproximation(z, lat0, lat1, z0, z1);
                iLon1 = linearApproximation(z, lon0, lon1, z0, z1);
                if (convexHullEdge != null) {
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat2, lon2, iLat0, iLon0, lat0, lon0);
                    addBoundaryLinesIfConvexHullMember(convexHullEdge, z, lat0, lon0, iLat1, iLon1, lat1, lon1);
                }
                break;
            default:
                System.err.println("***Contours: bad intersection case: " + intersectionCase);
                return null;
        }

        Debug.message("Contours2", "# intersection => ([" + iLat0 + "," + iLon0 + "] [" + iLat1 + "," + iLon1 + "])");
	if (Util.equals(iLat0, iLat1, 1E-10) && Util.equals(iLon0, iLon1, 1E-10))
	    return null;
        return new OMLine(iLat0, iLon0, iLat1, iLon1, OMGraphic.LINETYPE_STRAIGHT);
    }

    private static double linearApproximation(double z, double x1, double x2, double z1, double z2) {
        double x = x1 + (z - z1) * (x2 - x1) / (z2 - z1);
        // Debug.message("DataStorage", x1+" "+x2+" => "+x);
        return x;
    }
    private Set<OMPoly> convexHullTriangleSet;

    private void clearConvexHullTriangleSet() {
        convexHullTriangleSet = new HashSet<OMPoly>();
    }

    /** Return a DataPoint List representing an edge of the convex hull.
     *   As here, triangles contains that information, it is just taken from it.
     *   A set of triangle along the convex hull is built.
     */
    private Object getConvexHullEdge(OMPoly triangle) {
        Object ret = triangle.getAttribute(DelaunayTriangulationDataStorage.convexHullEdgeAttribute);
        if (ret != null) {
            convexHullTriangleSet.add(triangle);
        }
        return ret;
    }

    @SuppressWarnings("unchecked")
	private void addBoundaryLinesIfConvexHullMember(Object convexHullEdge, double contourValue, double lat0, double lon0, double iLat, double iLon, double lat1, double lon1) {
	// Check for (iLat, iLon) being on one extremity of the edge
	if (Util.equals(lat0, iLat, 1E-10) && Util.equals(lon0, iLon, 1E-10))
	    return;
	if (Util.equals(lat1, iLat, 1E-10) && Util.equals(lon1, iLon, 1E-10))
	    return;
	
	// (iLat, iLon) is on the interior of the line
        ArrayList<DataPoint> edge = (ArrayList<DataPoint>) convexHullEdge;
        // convexHullEdge is a list of points: at initialisation 2 points only,
        // then a point is inserted for each intersection
        DataPoint dp0 = edge.get(0);
        DataPoint dp1 = edge.get(edge.size() - 1);
        boolean convexHullMember = (dp0.equals(lon0, lat0) && dp1.equals(lon1, lat1)) || (dp1.equals(lon0, lat0) && dp0.equals(lon1, lat1));

        // System.out.println("? isConvexHullMember "+edge +" (["+ lat0+", "+ lon0+"] ["+ lat1+", "+ lon1 +"] => "+ convexHullMember);
        if (! convexHullMember) {
            return;
        }

	edge.add(new DataPoint(iLon, iLat, contourValue));
        Collections.sort(edge, new Comparator<DataPoint>() {
		
            public int compare(DataPoint dp1, DataPoint dp2) {
                int ret = (int) (dp2.coord(0) - dp1.coord(0));
                if (ret == 0) {
                    ret = (int) (dp2.coord(1) - dp1.coord(1));
                }
                return ret;
            }
        });
    }

    @SuppressWarnings("unchecked")
	private void addConvexHullLinesToDataObjects(ArrayList<DataObject> dataLineObjects, ArrayList<DataObject> dataObjects) {
	for(OMPoly triangle : convexHullTriangleSet) {
	    Object convexHullEdge = triangle.getAttribute(DelaunayTriangulationDataStorage.convexHullEdgeAttribute);
	    if (convexHullEdge == null) {
		System.err.println("Warning: no convex hull edge for "+ triangle);
		continue;
	    }
	    ArrayList<DataPoint> edge = (ArrayList<DataPoint>)convexHullEdge;
	    for(int i = 0; i < edge.size()-1; i++) {
		DataPoint dp0 = edge.get(i);
		DataPoint dp1 = edge.get(i+1);
		// Check for 0 length line
		if (Util.equals(dp0.coord(1), dp1.coord(1), 1E-10) && Util.equals(dp0.coord(0), dp1.coord(0), 1E-10))
		    continue;
		OMLine oml = new OMLine(
					dp0.coord(1), dp0.coord(0),
					dp1.coord(1), dp1.coord(0),
					OMGraphic.LINETYPE_STRAIGHT);
		DataObject contourData = new DataObject();
		OMDataLocation omdl = mkOMDataLocation(oml);
		double value;
		if (dp0.getData() instanceof DataObject)
		    value = Util.parseDouble(((DataObject) dp0.getData()).getData(getDataColumnName()));
		else
		    value = Util.parseDouble(dp0.getData());
		contourData.addData(getDataColumnName(), value);
		contourData.addOMDataLocation(getContourColumnName(), omdl);
		dataLineObjects.add(contourData);
		if (dataObjects != null)
		    dataObjects.add(contourData);
		Debug.message("Contours", "add line for "+contourData.getData(getDataColumnName())+ " " + omLineToString(oml));
	    }
	}
    }

    protected boolean setTable() {
        if (dataObjects == null) {
            System.err.println("[Contours] null data objects table");
            return false;
        }
	ArrayList<String> dataColumnNames = new ArrayList<String>();
	ArrayList<String> omColumnNames = new ArrayList<String>();
	if (withPolygons) {
	    dataColumnNames.add(getDataColumnName()+IntervalSuffix);
	    omColumnNames.add(getIsoPolyColumnName());
	}
	if (withLines) {
	    dataColumnNames.add(getDataColumnName());
	    omColumnNames.add(getContourColumnName());
	}
	if (addMesh) {
	    omColumnNames.add(meshColumnName);
	}
	dataTable.setTable(dataObjects, 
			   dataColumnNames, 
			   omColumnNames,  
			   new ArrayList<String>());
        ((DataTableStoragePanel) dtsp).fireStructureChanged();
        dsl.dataStorageEvent(this);
        return true;
    }

    public String[] getDataColumnNamesForTriangularGrid() {
        Set<String> names = new HashSet<String>();
        if (!isReady()) {
            return null;
        }
        Pattern p_012 = Pattern.compile(".*_[012]");
        for (String name : dataStorage.getDefaultDataTableModel().getDataColumnNames()) {
            if (p_012.matcher(name).matches()) {
                names.add(name.substring(0, name.length() - 2));
            }
        }

        // we should check for name_0, name_1, name_2 ...
        return names.toArray(new String[names.size()]);
    }

    public void setDataColumnName_012(String name) {
	dataColumnName = name;
        data0ColumnName = name + "_0";
        data1ColumnName = name + "_1";
        data2ColumnName = name + "_2";
    }

    private JPanel guiPanel;

    @Override
    public JComponent getGUI() {
        if (guiPanel == null) {
            guiPanel = new ContoursPanel(this);
        }
        return guiPanel;
    }

    public DataStorage removeItSelf() {
        dataStorage.dsl = dsl;
        dsl.dataStorageEvent(dataStorage);
        return dataStorage;
    }

    public void dataStorageEvent(DataStorage source) {
        Debug.message("DataStorage", "[Contours] dataStorageEvent <== " + source);
        if (isReady()) {
            Debug.message("DataStorage", "[Contours] data ready ***");
            initialize();
        }
    }

    public void setProperties(String prefix, java.util.Properties properties) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        super.setProperties(prefix, properties);

        prefix = PropUtils.getScopedPropertyPrefix(prefix) + contoursDSPropertyString + ".";
        String dsClassName = properties.getProperty(prefix + dataStorageClassPropertyString, "");
        if ("".equals(dsClassName)) {
            return;
        }
        dataStorage = Util.loadDataStorage(dsClassName, prefix, properties, this);

        geometryColumnName = properties.getProperty(prefix + geometryColumnNamePropertyString, "triangle");
        dataColumnName = properties.getProperty(prefix + dataColumnNamePropertyString, "");
        data0ColumnName = properties.getProperty(prefix + data0ColumnNamePropertyString, "COLOR_MAP_0");
        data1ColumnName = properties.getProperty(prefix + data1ColumnNamePropertyString, "COLOR_MAP_1");
        data2ColumnName = properties.getProperty(prefix + data2ColumnNamePropertyString, "COLOR_MAP_2");

        setWithLines(Boolean.parseBoolean(properties.getProperty(prefix + withLinesPropertyString, "true")));
        setWithPolygons(Boolean.parseBoolean(properties.getProperty(prefix + withPolysPropertyString, "true")));

        setIsoPolyColumnName(properties.getProperty(prefix + isoPolyColumnNamePropertyString, ""));
        setContourColumnName(properties.getProperty(prefix + contourColumnNamePropertyString, ""));

        String[] s = PropUtils.stringArrayFromProperties(properties, prefix + contoursValuesPropertyString, " ");
        if (s != null) {
            contoursValues = new ArrayList<Double>(s.length);
            for (int i = 0; i < s.length; i++) {
                contoursValues.add(Double.parseDouble(s[i]));
            }
        }
        initialize();
    }

    public Properties getProperties(String prefix, Properties properties) {
        properties = super.getProperties(prefix, properties);
        prefix = PropUtils.getScopedPropertyPrefix(prefix) + contoursDSPropertyString + ".";
        properties.put(prefix + geometryColumnNamePropertyString, geometryColumnName);

        if (dataStorage != null) {
            properties.put(prefix + dataStorageClassPropertyString, dataStorage.getClass().getCanonicalName());
            Debug.message("DataStorage", "ContoursDataStorage.getProperties() - DataStorage member is a: " + dataStorage.getClass().getCanonicalName());
            properties = dataStorage.getProperties(prefix, properties);
        }
        properties.put(prefix + geometryColumnNamePropertyString, geometryColumnName);
        properties.put(prefix + dataColumnNamePropertyString,getDataColumnName());
        properties.put(prefix + data0ColumnNamePropertyString, data0ColumnName);
        properties.put(prefix + data1ColumnNamePropertyString, data1ColumnName);
        properties.put(prefix + data2ColumnNamePropertyString, data2ColumnName);
        properties.put(prefix + withLinesPropertyString, String.valueOf(isWithLines()));
        properties.put(prefix + withPolysPropertyString, String.valueOf(isWithPolygons()));
        properties.put(prefix + getIsoPolyColumnName(),getIsoPolyColumnName());
        properties.put(prefix + getContourColumnName(),getContourColumnName());


        StringBuffer stringContoursValues = new StringBuffer();
        for (double contourValue : contoursValues) {
            stringContoursValues.append(" " + contourValue);
        }
        properties.put(prefix + contoursValuesPropertyString, stringContoursValues.toString());

        return properties;

    }

    public DataStorage getDataStorage() {
        return dataStorage;
    }

    public ArrayList<Double> getContoursValues() {
        return contoursValues;
    }

    /**
     * @return the withPolygons
     */
    public boolean isWithPolygons() {
        return withPolygons;
    }

    /**
     * @param withPolygons the withPolygons to set
     */
    public void setWithPolygons(boolean withPolygons) {
        this.withPolygons = withPolygons;
    }

    /**
     * @return the withLines
     */
    public boolean isWithLines() {
        return withLines;
    }

    /**
     * @param withLines the withLines to set
     */
    public void setWithLines(boolean withLines) {
        this.withLines = withLines;
    }

    /**
     * @return the contourColumnName
     */
    public String getContourColumnName() {
        return contourColumnName;
    }

    /**
     * @param contourColumnName the contourColumnName to set
     */
    public void setContourColumnName(String contourColumnName) {
        this.contourColumnName = contourColumnName;
    }

    /**
     * @return the isoPolyColumnName
     */
    public String getIsoPolyColumnName() {
        return isoPolyColumnName;
    }

    /**
     * @param isoPolyColumnName the isoPolyColumnName to set
     */
    public void setIsoPolyColumnName(String isoPolyColumnName) {
        this.isoPolyColumnName = isoPolyColumnName;
    }

    /**
     * @return the dataColumnName
     */
    public String getDataColumnName() {
        return dataColumnName;
    }

    public static String omLineToString(OMLine oml) {
	double[] lls = oml.getLL();
	return "["+lls[0]+", "+lls[1]+" - "+lls[2]+", "+lls[3]+"]";
    }
}
