// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.util.Debug;

import com.bbn.openmap.util.PropUtils;

import javax.swing.JComponent;
import javax.swing.JPanel;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.net.URL;
import java.io.File;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

import fr.unicaen.openmap.util.*;
import fr.unicaen.openmap.dataStorage.gui.DataTableModel;

import fr.unicaen.openmap.dataStorage.gml.out.GMLWriter;
import fr.unicaen.openmap.dataStorage.out.CSVWriter;

import com.bbn.openmap.PropertyConsumer;
import com.bbn.openmap.I18n;
import com.bbn.openmap.BasicI18n;
import com.bbn.openmap.omGraphics.OMGraphic;
import com.bbn.openmap.omGraphics.OMGraphicList;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer;
import java.util.Arrays;

/**
 * The <code>attributesXXXX</code> properties give the attributes to load in the datastorage, 
 * to display in the palette, and specify aliases for column names. 
 * Each name in the property is delimited with the regex listDelim (it's a Regex, <em>beware</em> of meta characters such as vertical bar |). 
 * If this list is not empty, all non named attibutes are skipped.
 * <p>
 * <pre>
 * ds.attributesToLoad=
 * ds.attributesToDisplay=
 * ds.attributesAliases=
 * ds.listDelim=,
 *</pre>
 *
 * @see DataStorageIterator
 */
public abstract class DataStorage implements PropertyConsumer {

    public static final String attributesToLoadPropertyString = "attributesToLoad";
    public static final String attributesToDisplayPropertyString = "attributesToDisplay";
    public static final String attributesAliasesPropertyString = "attributesAliases";
    public static final String listDelimPropertyString = "listDelim";
    public static final String warningPropertyString = "warning";
    protected DataStorageListener dsl;
    protected DataTableModel dataTableModel = null;
    /** DataTableStoragePanel */
    protected JPanel dtsp = null;
    /**
     * Layer class name, found unsing java.lang.reflect.
     */
    protected String classNameForDebug = Util.className(this);
    /**
     * Used to allow internationalization.
     */
    static protected I18n i18n = new BasicI18n();
    /**
     * name used by the GUI
     */
    protected String name = Util.className(this);

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public DataStorage(DataStorageListener dsl) {
        this.dsl = dsl;
        setName(i18n.get(this, "name", "Data Storage"));
    }

    public void setDataStorageListener(DataStorageListener dsl) {
        this.dsl = dsl;
    }

    public DataTableModel getDefaultDataTableModel() {
        return dataTableModel;
    }

    /**
     * Tells if the dataStorage has its data.
     */
    public abstract boolean isReady();

    /**
     * Get row number in this storage.
     */
    public abstract int getCount();

    /**
     * Random acccess to an element.
     */
    public abstract DataObject getElementAt(int i);

    public String getElementAtInHTML(int i) {
        DataObject data = getElementAt(i);
        if (dataAttributesNamesToDisplay != null && !dataAttributesNamesToDisplay.isEmpty()) {
            StringBuffer ret = new StringBuffer("<html><body>");
            ret.append("<table>");
            for (String attribute : dataAttributesNamesToDisplay) {
                // ret.append("<br><b>" + attribute + ":</b> " + data.getData(attribute).toString());
                String s;
                if (data.getData(attribute) instanceof Number) {
                    NumberFormat df = NumberFormat.getNumberInstance();
                    s = df.format((Number) data.getData(attribute));
                } else {
                    s = data.getData(attribute).toString();
                }
                ret.append("<tr><td><b>" + attribute + "</b></td><td> " + s + " </td></tr>");
            }
            ret.append("</table>");
            ret.append("</body></html>");
            return ret.toString();
        }
        return data.toHTML();
    }



    /**
     * Default Implementation of getGUI() method.
     * @return null
     */
    public JComponent getGUI() {
        // return new JLabel("No configuration for " + Util.className(this)+ "... we should be able to configure attributes list to load and to display !");
        return null;
    }

    /**
     *
     * Be careful, this this panel MUST be created by the implementation of DataStorage
     * only when Data are ready. If called before data are ready, this method return null;
     **/
    public JComponent getDataGUI() {
        return dtsp;
    }
    /**
     * Regex used as a property string list delimiter
     */
    protected String listDelim = ",";
    /**
     * List of attributes to load in this DataStorage
     */
    private ArrayList<String> dataAttributesNamesToLoad = null;
    /**
     * List of attributes to load in this DataStorage
     */
    private ArrayList<String> dataAttributesNamesToDisplay = null;
    /**
     * List of attributes to load in this DataStorage
     */
    private ArrayList<String> dataAttributesNamesAliases = null;

    /**
     * Tell if we should store this data attribute.
     * @return true if the name is contained in the list or if the list is null or empty.
     */
    public boolean isDataAttributeNameToLoad(String name) {
        return dataAttributesNamesToLoad == null || dataAttributesNamesToLoad.isEmpty() || dataAttributesNamesToLoad.contains(name);
    }

    /**
     * Returns the selected data attributes list that are to display.
     * It is the intersection of also columnNames, dataAttributesNamesToLoad and dataAttributesNamesToDisplay.
     * It is columnNames if dataAttributesNamesToLoad or dataAttributesNamesToDisplay are null or empty.
     */
    public ArrayList<String> getDataAttributesNamesToDisplay(ArrayList<String> columnNames) {
        ArrayList<String> dan;
        // we first restrict the list to loaded attributes
        dan = intersection(columnNames, dataAttributesNamesToLoad);
        // then to attributes to display
        dan = intersection(dan, dataAttributesNamesToDisplay);
        return dan;
    }

    public ArrayList<String> getDataAttributesNamesAliases() {
        return dataAttributesNamesAliases;
    }

    /**
     * Returns the selected data attributes list that are also columnNames
     * or columnNames if dataAttributesNamesToDisplay is null or empty
     * <br/>Prepends additionalName to the result list only if explicitely given
     * (special hack for adding esri record number in ShapeFileDataStorage).
     */
    public ArrayList<String> getDataAttributesNamesToDisplay(ArrayList<String> columnNames, String additionalName) {
        ArrayList<String> dan;
        // we first restrict the list to loaded attributes
        dan = intersection(columnNames, dataAttributesNamesToLoad);
        // then to attributes to display
        dan = intersection(dan, additionalName, dataAttributesNamesToDisplay);
        // mDebug.message("DataStorage" , "[" + classNameForDebug + "] getDataAttributesNamesToDisplay("+columnNames+", "+additionalName+") => "+dan);
        return dan;
    }

    /**
     * Default Implementation of setProperties() method. Should be always called in override definitions.
     */
    public void setProperties(String prefix, java.util.Properties properties) {

        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        //	setPropertyPrefix(prefix); //merdouille ?! A voir

        listDelim = properties.getProperty(prefix + listDelimPropertyString, ",");
        setWarning(Util.isTrue(properties.getProperty(prefix + warningPropertyString, "false")));

        // set the attributes names list
        dataAttributesNamesToLoad = UnicaenPropUtils.arrayListFromProperties(properties, prefix + attributesToLoadPropertyString, listDelim);
        dataAttributesNamesToDisplay = UnicaenPropUtils.arrayListFromProperties(properties, prefix + attributesToDisplayPropertyString, listDelim);
        dataAttributesNamesAliases = UnicaenPropUtils.arrayListFromProperties(properties, prefix + attributesAliasesPropertyString, listDelim);
        Debug.message("DataStorage", "[" + classNameForDebug + "]" + " attributes to load = " + dataAttributesNamesToLoad + " " + (dataAttributesNamesToLoad.size() == 0 ? " empty means no selection" : "") + " attributes to display = " + dataAttributesNamesToDisplay + " " + (dataAttributesNamesToDisplay.size() == 0 ? " empty means no selection" : "") + " attributes aliases = " + dataAttributesNamesAliases + " " + (dataAttributesNamesAliases.size() == 0 ? " empty means no selection" : ""));

    }

    /**
     * To save parameters for this DataStorage
     */
    public java.util.Properties getProperties(String prefix, Properties properties) {
        prefix = PropUtils.getScopedPropertyPrefix(prefix);
        //	setPropertyPrefix(prefix); //usage douteux pour le moment, mais nécessaire si on veut retrouver les propriétés, si elles existent (pour remplir les champs)

        properties.put(prefix + listDelimPropertyString, listDelim);
        properties.put(prefix + warningPropertyString, new Boolean(getWarning()).toString());

        UnicaenPropUtils.arrayListToPropertyString(properties, prefix + attributesToLoadPropertyString, dataAttributesNamesToLoad, listDelim);
        UnicaenPropUtils.arrayListToPropertyString(properties, prefix + attributesToDisplayPropertyString, dataAttributesNamesToDisplay, listDelim);
        UnicaenPropUtils.arrayListToPropertyString(properties, prefix + attributesAliasesPropertyString, dataAttributesNamesAliases, listDelim);

        return properties;
    }

    /**
     * From PropertyConsumer.
     * By default, this method provides properties by getProperties. So, it should be overrided in each subclass to provide exactly what is usefull.
     * @return properties
     */
    public java.util.Properties getPropertyInfo(Properties list) {
        if (list == null) {
            list = new java.util.Properties();
        }

        return list;
    }

    /**
     * To use with external getPropertyInfo(), if needed (to complete initPropertiesProperty by sample)
     * @return string contains property names used in getPropertyInfo
     */
    public String getInitPropertiesPropertyString() {
        return attributesToLoadPropertyString + " " + attributesToDisplayPropertyString + " " + attributesAliasesPropertyString + " " + listDelimPropertyString;
    }
    //To be continued... Conserver/restaurer le bon préfixe sans planter tout le reste. Pour le moment tout roule...  
    protected String prefix = " ";

    /**
     * From PropertyConsumer.
     */
    public void setPropertyPrefix(String prefix) {
        this.prefix = prefix;
    }

    /**
     * From PropertyConsumer.
     */
    public void setProperties(Properties properties) {
        setProperties(getPropertyPrefix(), properties);
    }

    /**
     * From PropertyConsumer.
     */
    public java.util.Properties getProperties(Properties properties) {
        return getProperties(getPropertyPrefix(), properties);
    }

    /**
     * From PropertyConsumer.
     */
    public String getPropertyPrefix() {
        return prefix;
    }

    /**
     * Computes the intersection of the given ArrayList with a special handling of null or empty lists.
     * Warning, this operator is not commutative !
     */
    private ArrayList<String> intersection(ArrayList<String> l1, ArrayList<String> l2) {
        ArrayList<String> inter = new ArrayList<String>();
        if (l1 == null || l1.isEmpty()) {
            return inter;
        }
        if (l2 == null || l2.isEmpty()) {
            return l1;
        }
        // really computes the mathematical set intersection
        for (String s : l2) {
            if (l1.contains(s)) {
                inter.add(s);
            } else {
                System.err.println("[" + classNameForDebug + "] ### Warning ### discard " + s + " from attributes list to load or display");
            }
        }
        return inter;
    }

    private ArrayList<String> intersection(ArrayList<String> l1, String additionalName, ArrayList<String> l2) {
        ArrayList<String> inter = new ArrayList<String>();
        if (l1 == null || l1.isEmpty()) {
            inter.add(0, additionalName);
            return inter;
        }
        if (l2 == null || l2.isEmpty()) {
            inter.addAll(l1);
            inter.add(0, additionalName);
            return inter;
        }

        // really computes the mathematical set intersection
        for (String s : l2) {
            if (l1.contains(s) || s.equals(additionalName)) {
                inter.add(s);
            } else {
                System.err.println("[" + classNameForDebug + "] ### Warning ### discard " + s + " from attributes list to load or display");
            }
        }
        return inter;
    }

    public void write(CSVWriter writer) {
	writer.startWrite();
	for (String col : dataTableModel.getDataColumnNames())
	    writer.writeData(col);

	// list of exportable centroids (not already in data)
	List<String> centroidColunmNames = new ArrayList<String>();
	for (String col : dataTableModel.getOMColumnNames()) {
	    int i = col.indexOf("/");
	    if (i == -1)
		continue;
	    String lonName = col.substring(0,i);
	    String latName = col.substring(i+1);
	    if (dataTableModel.getIndexForColumnName(lonName) != -1 && 
		dataTableModel.getIndexForColumnName(latName) !=-1)
		continue; // already in data
	    centroidColunmNames.add(col);
	    writer.writeData(lonName);
	    writer.writeData(latName);
	}
	writer.newLine();

       for (int i = 0; i < getCount(); i++) {
	   DataObject dataObject = getElementAt(i);
	   for (String col : dataTableModel.getDataColumnNames())
	       writer.writeData(dataObject.getData(col));
	   for (String col : centroidColunmNames) {
	       OMDataLocation omdl = dataObject.getOMDataLocation(col);
	       Centroid centroid = omdl == null ? null : omdl.getCentroid();
	       writer.writeData(centroid == null ? "" : centroid.getLon());
	       writer.writeData(centroid == null ? "" : centroid.getLat());
	   }
	   writer.newLine();
       }
       writer.close();
    }

    public void write(GMLWriter writer) {
        String originType = "GMLDataStorage";
        writer.startWrite(originType);
        for (int i = 0; i < getCount(); i++) {
            writer.startOMPGFeature();
            DataObject dObj = getElementAt(i);
            for (int j = 0; j < dataTableModel.getOMColumnCount(); j++) {
                String col = dataTableModel.getOMColumnName(j);
                OMDataLocation omdl = dObj.getOMDataLocation(col);
                if (omdl != null) {
                    writer.startGeometryProperty();
                    if (omdl.getOMGraphic() != null) {
                        writer.writeCentroidAndBoundingBox(omdl);
                        writer.startMultiPolygon();
                        writer.writeGMLElement(omdl.getOMGraphic());
                        writer.endMultiPolygon();
                    } else {
                        writer.writeCentroid(omdl);
                    }
                    writer.endGeometryProperty();
                }
                /*
                 * if (omg instanceof OMGraphicList) { OMGraphicList omgl =
                 * (OMGraphicList) omg; Iterator it = omgl.iterator(); Object o =
                 * null; while (it.hasNext()) { o = it.next();
                 */
                // writer.addDataLocation(k,omdl);
                // }
            }

            for (int j = 0; j < dataTableModel.getColumnCount(); j++) {
                String col = dataTableModel.getColumnName(j);
                Object obj = dObj.getData(col);
                if (obj != null) {
                    String type;
                    if (obj instanceof Number) {
                        type = "decimal";
                    } else {
                        type = "string";
                    }

                    writer.addData(col, dObj.getData(col).toString(), type);
                }
            }
            writer.endOMPGFeature();
        }
	writer.closeDocument();
    }
    private boolean warning = false;

    public boolean getWarning() {
        return warning;
    }

    public void setWarning(boolean b) {
        warning = b;
    }

    public boolean isAnExistingColumnName(String col) {
        return Arrays.asList(dataTableModel.getDataColumnNames()).contains(col) ||
                Arrays.asList(dataTableModel.getOMColumnNames()).contains(col);
    }

    /** 
     * Utility function that strip any directory prefix of the file name if it is in the CLASSPATH
     */
    public static String stripFileNameWithClassPathDirectories(String oldpath) {
        if (oldpath == null || "".equals(oldpath))
            return oldpath;
        String path = oldpath;
        Debug.message("path", "## " + oldpath);
        String classPath = System.getenv("CLASSPATH");
        if (classPath == null || "".equals(classPath))
            return path;
        try {
            URL url = new URL(path);
            if ("file".equals(url.getProtocol())) {
                path = url.getFile();
            } else {
                // we don't change URL that are not using the file protocol.
                return path;
            }
        } catch (java.net.MalformedURLException e) {
        }
        // we have a true file path name.
        try {
            File file = new File(path).getCanonicalFile();
            path = file.getPath();
        } catch (java.io.IOException e) {
            return path;
        }

        // match against classPathDirectories
        for (String dir : getClassPathDirectories()) {
            //System.out.println("Compare "+dir + " to "+ path);
            if (path.startsWith(dir))
                return path.substring(dir.length() + 1);
        }
        Debug.message("path", "### " + oldpath + " => " + path);
        return path;
    }
    /** List of directories given in the CLASSPATH environnement variable */
    private static List<String> classPathDirectories;

    public static List<String> getClassPathDirectories() {
        if (classPathDirectories != null)
            return classPathDirectories;
        classPathDirectories = new ArrayList<String>();
        String classPath = System.getenv("CLASSPATH");
        if (classPath == null || "".equals(classPath))
            return classPathDirectories;
        Pattern dotjarPattern = Pattern.compile("\\.jar$");
        for (String p : classPath.split(":")) {
            Matcher m = dotjarPattern.matcher(p);
            if (m.matches()) // jar file
                continue;
            // is it a valid directory ?
            try {
                File file = new File(p).getCanonicalFile();
                if (file.isDirectory()) {
                    String dirpath = file.getPath();
                    if (dirpath != null && !"".equals(dirpath))
                        classPathDirectories.add(dirpath);
                }
            } catch (java.io.IOException e) {
            }
        }
        Collections.sort(classPathDirectories, new Comparator<String>() {

            public int compare(String s1, String s2) {
                return s1.length() - s2.length();
            }
        });
        return classPathDirectories;
    }

    /**
     * Returns the list of the DataStorage of the layer
     * excluding only HashedDataStorage (@seee DualDataStorage.getSecondDataStorage())
     */
    public static List<DataStorage> getDataStorageList(fr.unicaen.openmap.layer.thematicDataDisplayLayer.ThematicDataDisplayLayer layer) {
        if (layer == null || layer.getDataStorageIterator() == null)
            return new ArrayList<DataStorage>();
        return layer.getDataStorageIterator().getDataStorage().getDataStorageList();
    }

    /**
     * Returns the list of the sub DataStorage of this DataStorage including itself,
     * excluding only HashedDataStorage (@seee DualDataStorage.getSecondDataStorage())
     * Could be rewrited in a more OO programming style if we add to DataStorage the method : getDataStorage() and getSecondDataStorage().
     */
    public List<DataStorage> getDataStorageList() {
        List<DataStorage> list = new ArrayList<DataStorage>();
        list.add(this);
        if (this instanceof RemovableDataStorage) {
            if (this instanceof DualDataStorage)
                list.addAll(((DualDataStorage) this).getSecondDataStorage().getDataStorageList());
            list.addAll(((RemovableDataStorage) this).getDataStorage().getDataStorageList());
        }
        return list;
    }

    /**
     * Returns the list of the sub DataStorage of this DataStorage that are instance of AnimationDataStorage. 
     * May include this if it is itself an AnimationDataStorage.
     * @see getDataStorageList()
     */
    public List<MonoValuedFilterDataStorage> getAnimationDataStorageList() {
        List<MonoValuedFilterDataStorage> list = new ArrayList<MonoValuedFilterDataStorage>();
        if (this instanceof MonoValuedFilterDataStorage)
            list.add((MonoValuedFilterDataStorage) this);
        if (this instanceof RemovableDataStorage) {
            if (this instanceof DualDataStorage)
                list.addAll(((DualDataStorage) this).getSecondDataStorage().getAnimationDataStorageList());
            list.addAll(((RemovableDataStorage) this).getDataStorage().getAnimationDataStorageList());
        }
        return list;
    }

    public ThematicDataDisplayLayer getLayer() {
        if (dsl instanceof ThematicDataDisplayLayer)
            return (ThematicDataDisplayLayer) dsl;
	else if (dsl instanceof DataStorage)
	    return ((DataStorage) dsl).getLayer();
	return null;
    }
}
