// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.dataStorage;

import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.proj.ProjMath;

import java.util.Iterator;

/**
 * Stores a data location made up with a OMGraphic and optionally its centroid 
 * and bounding box. centroid and bounding box may be null. 
 */
public class OMDataLocation {

    private OMGraphic omGraphic = null;
    private Centroid centroid = null;
    private BoundingBox boundingBox = null;

    /**
     * Constructor
     * @param omGraphic an OMGraphic (generally a shape or a line)
     * @param centroid its associated centroid defined in lat,lon
     * @param boundingBox the boundingBox box of the OMGraphic. 
     */
    public OMDataLocation(OMGraphic omGraphic, Centroid centroid, BoundingBox boundingBox){
	this.omGraphic = omGraphic;
	this.centroid  = centroid;
	this.boundingBox  = boundingBox;
    }
    
    /**
     * Simple constructor.
     * centroid and boundingBox will be null
     * @param omGraphic an OMGraphic (generally a shape or a line)
     */
    public OMDataLocation(OMGraphic omGraphic) { this(omGraphic, null, null); }
    
    /**
     * We need this constructor to set a completely empty OMDataLocation.
     *
     */
    public OMDataLocation () { }

    /**
     * Get the Centroid object associated to OMGraphic.
     * @return a Centroid or null
     */
    public Centroid getCentroid(){
	return centroid;
    }

    /**
     * Get the BoundingBox object associated to OMGraphic.
     * @return a BoundingBox or null.
     */
    public BoundingBox getBoundingBox(){
	return boundingBox;
    }

    /**
     * Set the Centroid object associated to OMGraphic.
     */
    public void setCentroid(Centroid c){
	centroid = c;
    }

    /**
     * Set the BoundingBox object associated to OMGraphic.
     */
    public void setBoundingBox(BoundingBox b){
	boundingBox = b;
    }

    /**
     * Get the OMGraphic
     * @return OMGraphic
     */
     public OMGraphic getOMGraphic(){
	return omGraphic;
    }
    /**
     *
     */
    public void setOMGraphic(OMGraphic omGraphic) {
	this.omGraphic = omGraphic;
    }

    public String toString() {
	return "{"+omGraphic+", "+centroid+", "+boundingBox+"}";
    }

    /**
     * Try to compute "intelligently" a centroid of a list of polygons 
     * as the centroid of the largest polygon. It has only sense for RENDERTYPE_LATLON omgraphics.
     * <br>The fallback crude method is to return the center of the bounding box.
     * <p>The omGraphic and the bounding box must be set before to call this method.
     */
    public void initCentroidFromOMGraphic() {
	// search for the largest area polygon in the graphic list
//	double maxArea = Double.MIN_VALUE;
//	OMPoly omp = largestOMPoly(omGraphic, maxArea);
//	System.out.println("initCentroidFromOMGraphic " + omp);
//	if (omp != null)
//	    setCentroid(calculateCentroid(omp));
//	else // fall back to center of bounding box
	    setCentroid(new Centroid((boundingBox.getUpperLeftLon() + boundingBox.getLowerRightLon())/2.0,
				     (boundingBox.getUpperLeftLat() + boundingBox.getLowerRightLat())/2.0));
    }
    
    private OMPoly largestOMPoly(OMGraphic omg, Double maxArea) {
	if (omg instanceof OMGraphicList) {
	    OMPoly ret = null;
	    OMGraphicList omgl = (OMGraphicList) omg;
	    for (Iterator it = omgl.iterator(); it.hasNext();) {
		OMGraphic g = (OMGraphic) it.next();
		OMPoly lomp = largestOMPoly(g, maxArea);
		// System.out.println(" largestOMPoly " +lomp);
		if (lomp != null)
		    ret = lomp;
	    }
	    return ret;
	} else if (omg instanceof OMPoly) {
	    OMPoly omp = (OMPoly)omg;
	    //	    System.out.println (omp.getCoordMode());
	    //	    if (omp.getCoordMode() != OMGraphicConstants.RENDERTYPE_LATLON)
	    //		return null;
	    // Funny: CoordMode appears to be always RENDERTYPE_UNKNOWN !!!
	    double area = calculateArea(omp);
	    // System.out.println("  area " + area);
	    if (area > maxArea) {
		maxArea = area;
		return omp;
	    }
	    return null;
	}
	// System.out.println("## "+omg.getClass());
	return null;
    }

   /**
     * Calculate the area of the poly. Algorithm used is
     * from some australian astronomy website =)
     * http://astronomy.swin.edu.au/~pbourke/geometry/polyarea
     * Idea from @see com.bbn..openmap.omGraphics.labeled.LabeledOMPoly
     */
    private double calculateArea(OMPoly omp) {
        double area = 0.0;
	double rawllpts[] = omp.getRawllpts();
        for (int i = 0; i < rawllpts.length; i+=2) {
            int j = (i + 2) % rawllpts.length;
            area += rawllpts[i] * rawllpts[j+1]; // xpts[i] * ypts[j];
            area -= rawllpts[i+1] * rawllpts[j]; // ypts[i] * xpts[j];
        }
      
        return area / 2.0;
    }
    /**
     * Calculate a centroid. 
     * Idea of the algorithm : the centroid is the baricenter of the edges weighted by their length.  
     */
    private Centroid calculateCentroid(OMPoly omp) {
	double clat = 0;
        double clon = 0;
	double tlength = 0;

	double rawllpts[] = omp.getRawllpts();

        for (int i = 0; i < rawllpts.length; i+=2) {
            int j = (i + 2) % rawllpts.length;
	    double lat = rawllpts[j]-rawllpts[i];
	    double lon = rawllpts[j+1]-rawllpts[i+1];
	    
	    double length = Math.sqrt(lat*lat+lon*lon);
	    clat += length*(rawllpts[j]+rawllpts[i])/2.0;
	    clon += length*(rawllpts[j+1]+rawllpts[i+1])/2.0;
	    tlength += length;
	}
        clat /= tlength;
        clon /= tlength;

	return new Centroid(ProjMath.radToDeg(clat), ProjMath.radToDeg(clon));
    }
	
}
