package fr.unicaen.openmap.layer;

import com.bbn.openmap.MapBean;
import com.bbn.openmap.MouseDelegator;
import com.bbn.openmap.gui.DefaultHelpMenu;
import java.util.Properties;
import com.bbn.openmap.layer.OMGraphicHandlerLayer;
import com.bbn.openmap.omGraphics.*;
import com.bbn.openmap.proj.*;
import com.bbn.openmap.proj.coords.LatLonPoint;
import com.bbn.openmap.util.*;
import com.bbn.openmap.util.propertyEditor.Inspector;

import fr.unicaen.openmap.util.Util;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import javax.imageio.ImageIO;
import java.awt.image.*;
import java.awt.geom.Point2D;
import java.beans.beancontext.BeanContext;
import javax.swing.JPanel;
import java.beans.PropertyEditorSupport;
import com.bbn.openmap.BasicI18n;
import com.bbn.openmap.I18n;

/**
 * Sample implementation of what we need to display a raster image of the Earth...
 */
public class RasterLayer extends OMGraphicHandlerLayer {
 
    /**
     * Use I18N mecanism to allow internationalization.
     * Use I18N.properties file or I18N_fr_FR.properties for french version
     **/
    protected I18n i18n = new BasicI18n();
    
    /**
     * Layer class name, found unsing java.lang.reflect.
     */
    protected String classNameForDebug = Util.className(this);
    
    static public final String imageFileNamePropertyString = "imageFileName";
    /** 
     * lat/lon bounding box of the image 
     * lonMin, lonMax, latMin, latMax are the lat/lon corresponding to 
     * the minX, maxX, minY and maxY of the image
     */
    static public final String lonMinPropertyString = "lonMin";
    static public final String lonMaxPropertyString = "lonMax";
    static public final String latMinPropertyString = "latMin";
    static public final String latMaxPropertyString = "latMax";

    protected String imageFileName = "";
    protected double lonMin = -180 , lonMax = +180, latMin = 90, latMax = -90;
 
    public RasterLayer(){
        Debug.message(classNameForDebug, this + " new RasterLayer()");
    }
    
    /**
     * Sets the properties for the Layer in order to initialize.
     * @param prefix string prefix used in the properties file for this layer
     * @param properties the properties set in the properties file
     */
    public void setProperties(String prefix, Properties properties) {
        super.setProperties(prefix, properties);

        prefix = PropUtils.getScopedPropertyPrefix(prefix);

	imageFileName = properties.getProperty(prefix + imageFileNamePropertyString);
	lonMin = 
	    PropUtils.doubleFromProperties(properties, prefix + lonMinPropertyString, -180);
	lonMax = 
	    PropUtils.doubleFromProperties(properties, prefix + lonMaxPropertyString, +180);
	latMin = 
	    PropUtils.doubleFromProperties(properties, prefix + latMinPropertyString,  +90);
	latMax = 
	    PropUtils.doubleFromProperties(properties, prefix + latMaxPropertyString, -90);

        Debug.message(classNameForDebug, this + " setProperties(...) : RasterLayer "+imageFileName);
        
   }
    
    /**
     * PropertyConsumer method, to fill in a Properties object, reflecting the current values of the layer
     * in order to save it to a file.
     * @param properties the properties set in  the properties file
     * @return Properties.
     */
    public Properties getProperties(Properties properties) {
        properties = super.getProperties(properties);
        Debug.message(classNameForDebug, this + " RasterLayer getProperties(...)"+properties);
        
        String prefix = PropUtils.getScopedPropertyPrefix(this);
	
	properties.put(prefix + imageFileNamePropertyString, imageFileName);

	properties.put(prefix + lonMinPropertyString, String.valueOf(lonMin));
	properties.put(prefix + lonMaxPropertyString, String.valueOf(lonMax));
	properties.put(prefix + latMinPropertyString, String.valueOf(latMin));
	properties.put(prefix + latMaxPropertyString, String.valueOf(lonMax));

        return properties;
    }

   /**
     * Provides the essential information in order to build a GUI
     */
    public java.util.Properties getPropertyInfo(Properties info) {

        info = super.getPropertyInfo(info);

	info.put(initPropertiesProperty, imageFileNamePropertyString+" "+lonMinPropertyString+" "+lonMaxPropertyString+" "+latMinPropertyString+" "+latMaxPropertyString);

	PropUtils.setI18NPropertyInfo(i18n,
				      info,
				      RasterLayer.class,
				      imageFileNamePropertyString,
				      "Image file name",
				      "the raster image",
				      "com.bbn.openmap.util.propertyEditor.FilePropertyEditor"
				      );
	PropUtils.setI18NPropertyInfo(i18n,
				      info,
				      RasterLayer.class,
				      lonMinPropertyString,
				      "lon min",
				      "longitude corresponding to the minX of the raster image",
				      null //default editor
				      );
	PropUtils.setI18NPropertyInfo(i18n,
				      info,
				      RasterLayer.class,
				      lonMaxPropertyString,
				      "lon max",
				      "longitude corresponding to the maxX of the raster image",
				      null //default editor
				      );
	PropUtils.setI18NPropertyInfo(i18n,
				      info,
				      RasterLayer.class,
				      latMinPropertyString,
				      "lat min",
				      "latitude corresponding to the minY of the raster image",
				      null //default editor
				      );
	PropUtils.setI18NPropertyInfo(i18n,
				      info,
				      RasterLayer.class,
				      latMaxPropertyString,
				      "lat max",
				      "latitude corresponding to the maxY of the raster image",
				      null //default editor
				      );
	return info;
    }

    
    /**
     * Sample function to render the layer as a string. You don't need
     * to override this method to match the name of your subclass since
     * we use java.lang.reflect to find the class name.
     */
    public String toString() { return "[" + classNameForDebug + ":" + getName() + "]"; }
    
    /**
     * Basic sample implementation. You should override this
     * function to match your layer interfaces need.
     */
    public Component getGUI() {
	JPanel panel = new JPanel();
	final Inspector inspector = new Inspector();
        panel.add(inspector.createPropertyGUI(this));

        return panel;
    }
    
    /**
     * Used to redraw all elements.
     */
    public void refresh() {
        Debug.message(classNameForDebug, this + " refresh()");
        // use Swing Worker ...
        doPrepare();        
    }
    
    /**
     * Create the OMGraphics to be displayed by the MapBean.
     *
     * @return OMGraphicList
     */
    public synchronized OMGraphicList prepare() {
        Projection proj = getProjection();
	if (proj == null) {
            Debug.message(classNameForDebug, this + " no projection yet, skipping prepare()");
            return null;
        }
        OMGraphicList omgl = null;
        setList(null);
        Debug.message(classNameForDebug, this + " prepare() used !");

        omgl = generateGraphics();
	omgl.generate(proj);

        setList(omgl);
	Debug.message(classNameForDebug, this + " " + omgl.size() + " OMGraphics");
	return omgl;
    }

    protected BufferedImage image = null;

    protected void initImage() {
	if (image != null)
	    return;
	try {
	    image = ImageIO.read(Util.open(imageFileName));
	} catch (java.io.IOException ioe) {
	    System.err.println("RasterLayer: can't open "+imageFileName);
	}
    }
    

    protected Projection proj = null;
    private Point2D center;
    private Point2D p; // used by proj.inverse to avoid realloc
    protected int getPixel(int xx, int yy) {
	p = proj.inverse(xx, yy, p);
	double lon = p.getX();
	double lat = p.getY();
	if (p.equals(center)) // inverse retruns center in case of point out of the map
	    return 0; // black
	try {
	    double x = image.getMinX() + image.getWidth() * (lon - lonMin) / (lonMax - lonMin);
	    int x1 = (int)Math.floor(x);
	    int x2 = (int)Math.ceil(x);
	    double y = image.getMinY() + image.getHeight() * (lat - latMin) / (latMax - latMin);
	    int y1 = (int)Math.floor(y);
	    int y2 = (int)Math.ceil(y);
	    
	    int x1y1 = image.getRGB(x1, y1);
	    int x2y1 = image.getRGB(x2, y1);
	    int x1y2 = image.getRGB(x1, y2);
	    int x2y2 = image.getRGB(x2, y2);
	    
	    int alpha = 
		bilinearApprox(getAlpha(x1y1), getAlpha(x2y1), getAlpha(x1y2), getAlpha(x2y2), 
			       x1, x2, y1, y2, x,  y);
	    int red = 
		bilinearApprox(getRed(x1y1), getRed(x2y1), getRed(x1y2), getRed(x2y2), 
			       x1, x2, y1, y2, x,  y);
	    int green = 
		bilinearApprox(getGreen(x1y1), getGreen(x2y1), getGreen(x1y2), getGreen(x2y2), 
			       x1, x2, y1, y2, x,  y);
	    int blue = 
		bilinearApprox(getBlue(x1y1), getBlue(x2y1), getBlue(x1y2), getBlue(x2y2), 
			       x1, x2, y1, y2, x,  y);

	    return  (alpha << 24) | (red << 16) | (green << 8) | blue; 
	} catch (ArrayIndexOutOfBoundsException e) {
		return 0; // black
	}
    }

    private static int getAlpha(int rgb) { return (rgb >> 24) & 0xff; }
    private static int getRed(int rgb) { return (rgb >> 16) & 0xff; }
    private static int getGreen(int rgb) { return (rgb >> 8) & 0xff; }
    private static int getBlue(int rgb) { return rgb & 0xff; }
    
    private static int bilinearApprox(int x1y1, int x2y1, int x1y2, int x2y2, int x1, int x2, int y1, int y2, double x, double y) {
	return (int)Math.round(((x1y1*(x2-x)+x2y1*(x-x1))*(y2-y)
				 + (x1y2*(x2-x)+x2y2*(x-x1))*(y-y1))
				/ (x2-x1)*(y2-y1));
    }
//    protected int filter(int i){
//        return i + 80 << 8;
//    }

    public OMGraphicList generateGraphics() {
	
	initImage();
	if (image == null) 
	    return null;

	proj = getProjection();
	if (proj == null) {
	    System.err.println("RasterLayer: no projection has been set");
	    return null;
	}

	int w = proj.getWidth();
	int h = proj.getHeight();
	center = proj.getCenter();

	BufferedImage bim = new BufferedImage (w, h, BufferedImage.TYPE_INT_ARGB);
	for(int x = 0; x < w; x++)
	    for (int y = 0; y < h; y++)
		bim.setRGB(x, y, getPixel(x, y));
	    
	OMRaster omr = new OMRaster(0, 0, bim);
	OMGraphicList omgl = new OMGraphicList();
	omgl.add(omr);

	return omgl;
    }
}
