// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Properties;
import java.lang.reflect.Constructor;
import javax.swing.*;

import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.ColorFactory;

import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.ColorScalePanel2;

/**
 * ColorScale is a fixed number of colors.
 * <p>May be used to assign a color to a graphical object, given a value.
 * <p>Default colors are:
 * <pre>c9ffb300 c9ec93c1 c974c476 c980a6ce c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5</pre> 
 * <p>May be set by properties. Ex.:<br>
 * <pre>
 * prefix.colorScale = c9ffb300 c9ec93c1 c974c476 c980a6ce c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5
 * </pre>
 * or
 * <pre>
 * prefix.colorScale = c9ff0000 c9dd2200 c9bb4400 c9996600 c9778800 C955aa00 c933cc00 c901dd00 c900ff00
 * </pre>
 */
public class ColorScale {

    private ArrayList<Color> colorScale = new ArrayList<Color>();
    private int colorNumber = 0;
    private boolean reverse = false;
    public static final String colorScalePropertyString = "colorScale";
    public static final String defaultColorScaleInHex =
            "ffffb300 ffec93c1 ff74c476 ff80a6ce";// c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5
    private String displayName = "";

    public ColorScale() {
        initFromColorScaleInHex(defaultColorScaleInHex);
    }

    /**
     * Ctor from a string plus a name
     */
    public ColorScale(String colorScaleInHex, String displayName) {
        this.displayName = displayName;
        initFromColorScaleInHex(colorScaleInHex);
    }

    /**
     * Used to display in a JComboBox.
     */
    public String toString() {
        return displayName;
    }

    /**
     * Use this method for tracing instead of toString
     */
    public String debug() {
        String s = "ColorScale (" + colorNumber + ") ";
        for (Color c : colorScale) {
            s += c + " ";
        }
        return s;
    }

    public void setProperties(String prefix, Properties properties) {
        String colorScaleInHex =
                properties.getProperty(prefix + colorScalePropertyString, defaultColorScaleInHex);
        initFromColorScaleInHex(colorScaleInHex);
    }

    public void initFromColorScaleInHex(String colorScaleInHex) {
        colorScale = new ArrayList<Color>();
        colorNumber = 0;
        String colorScaleTabInHex[] = colorScaleInHex.split("[ \t]+");
        for (int i = 0; i < colorScaleTabInHex.length; i++) {
            if ("".equals(colorScaleTabInHex[i])) {
                continue;
            }
            colorScale.add(ColorFactory.parseColor(colorScaleTabInHex[i]));
            displayName += ColorFactory.parseColor(colorScaleTabInHex[i]) + " ";
            colorNumber++;
        }
    }

    public Properties getProperties(String prefix, Properties properties) {
        String colorScaleInHex = "";
        for (int i = 0; i < colorNumber; i++) {
            colorScaleInHex += " " + ColorFactory.getHexColorString(colorScale.get(i));
        }
        properties.put(prefix + colorScalePropertyString, colorScaleInHex);
        return properties;
    }

    public Color getColor(int index) {
        return index >= 0 && index < colorScale.size()
                ? colorScale.get(index)
                : Color.white; // should return null in order to trap bad programming !-:
    }

    public void setColor(int index, Color color) {
        if (index < 0 || index >= colorScale.size()) {
            System.err.println("ColorScale : error index out of bounds : " + index + ", " + colorScale.size());
            return;
        }
        // set the color
        colorScale.set(index, color);
    }

    public int getColorNumber() {
        return colorNumber; /* colorScale.size(); */ }

    public void setColorNumber(int nb) {
        for (int i = colorScale.size(); i < nb; i++) {
            colorScale.add(new Color((int) (Math.random() * 255), (int) (Math.random() * 255), (int) (Math.random() * 255)));
        }
        //colorScale.add(Color.white);
        colorNumber = nb;
        // System.out.println("### ColorScale.setColorNumber("+nb+") => colorScale.size() = " + colorScale.size()+ " colorScale "+ this);
    }

    /** @deprecated see derived classes of ColorScalePanel */
    public JComponent getGUI() {
        return new ColorScalePanel2(this);
    }

    /**
     * init a shade color map given the two extremities colors
     */
    public void doShade() {
        Color min = colorScale.get(0);
        Color max = colorScale.get(colorNumber - 1);
        for (int i = 1; i < colorNumber - 1; i++) {
            int r = min.getRed() + (int) (i * (max.getRed() - min.getRed()) / colorNumber);
            int g = min.getGreen() + (int) (i * (max.getGreen() - min.getGreen()) / colorNumber);
            int b = min.getBlue() + (int) (i * (max.getBlue() - min.getBlue()) / colorNumber);
            int a = min.getAlpha() + (int) (i * (max.getAlpha() - min.getAlpha()) / colorNumber);
            setColor(i, new Color(r, g, b, a));
        }
    }

    /**
     * Copy the colors from ColorScale c.
     */
    public void copy(ColorScale c) {
        //setColorNumber(c.getColorNumber());
        for (int i = 0; i < colorScale.size(); i++) {
            setColor(i, c.getColor(i));
        }

    }

    public void reverseCopy(ColorScale c) {
        //setColorNumber(c.getColorNumber());
        int nb = getColorNumber() - 1;
        for (int i = 0; i < getColorNumber(); i++, nb--) {
            setColor(nb, c.getColor(i));
        }
    }

    public void setReverse(boolean reverse) {
        this.reverse = reverse;
    }

    public boolean getReverse() {
        return reverse;
    }
}
