// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.layer.thematicDataDisplayLayer.util;

import java.awt.Color;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Properties;
import java.lang.reflect.Constructor;
import javax.swing.*;
import javax.swing.event.*;

import com.bbn.openmap.util.PropUtils;
import com.bbn.openmap.util.Debug;
import com.bbn.openmap.util.ColorFactory;
import com.bbn.openmap.omGraphics.*;

import fr.unicaen.openmap.util.classification.*;
import fr.unicaen.openmap.dataStorage.*;
import fr.unicaen.openmap.layer.thematicDataDisplayLayer.gui.*;

/**
 * A Colorizer is made of a color scale (i.e. a fixed number of colors) and a
 * classification.
 * <p>
 * May be used to assign a color to a graphical object, given a value.
 * <p>
 * Default colors are:
 * 
 * <pre>
 * c9ffb300 c9ec93c1 c974c476 c980a6ce c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5
 * </pre>
 * 
 * <p>
 * May be set by properties. Ex.:<br>
 * 
 * <pre>
 * prefix.colorScale = c9ffb300 c9ec93c1 c974c476 c980a6ce c9ffff4d c9f4c6de c9a6dba0 c9bfd1e5
 * prefix.classificationClassName = fr.unicaen.openmap.util.classification.NestedMeans
 * prefix.nbClasses = 4
 * </pre>
 */
public class Colorizer {

    private ColorScale colorScale = new ColorScale();
    // private Classification classif = null;
    // private Classification classif = new Quantile(null,
    // colorScale.getColorNumber());
    // private Classification classif = new StepClassification(null,
    // colorScale.getColorNumber()); //confusion
    private Classification classif;// = new StepClassification(new
    // CustomizedClassification(null,
    // colorScale.getColorNumber()),
    // colorScale.getColorNumber());
    protected String stringNumberFormat = "%5.2f";
    public static final String classificationClassNamePropertyString = "classificationClassName";
    public static final String nbClassesPropertyString = "nbClasses";
    /**
     * For CustomizedClassification.
     */
    public static final String BREAKS_PROPERTY_NAME = "breaks";
    // Event programming for classification change.
    // Hand made because i don't manage to use EventListenerList.
    private HashSet<ClassificationListener> listenerList = new HashSet<ClassificationListener>();

    public void addClassificationListener(ClassificationListener l) {
        listenerList.add(l);
    }

    public void removeClassificationListener(ClassificationListener l) {
        listenerList.remove(l);
    }

    protected void fireClassificationChanged() {
        int i = 0;
        for (ClassificationListener l : listenerList) {
            Debug.message("Colorizer", "##Colorizer.fireClassificationChanged " + (i++) + " => " + l);
            l.classificationChanged(this);
        }
    }

    public void setProperties(String prefix, Properties properties) {
        colorScale.setProperties(prefix, properties);
        String classificationClassName =
                properties.getProperty(prefix + classificationClassNamePropertyString,
                null);
        int nbClasses =
                PropUtils.intFromProperties(properties,
                prefix + nbClassesPropertyString,
                4);
        if (classificationClassName != null) {
            setClassification(classificationClassName, nbClasses);
        }
        // Custom mode
        String[] s = PropUtils.stringArrayFromProperties(properties, prefix + BREAKS_PROPERTY_NAME, "\t");
        if (s != null) {
            double[] breaks = new double[s.length];
            for (int i = 0; i < s.length; i++) {
                breaks[i] = Double.parseDouble(s[i]);
            }
            if (classif instanceof CustomizedClassification) {
                ((CustomizedClassification) classif).setBreaks(breaks);
            }
        }

        colorScale.setColorNumber(nbClasses);
    }

    public void setClassification(String classificationClassName, int nbClasses) {
        // System.out.println("#Colorizer.setClassification("+classificationClassName+",
        // "+nbClasses+")");
        Classification oldClassif = classif;
        try {
            Class cls = Class.forName(classificationClassName);
            Constructor[] constructors = cls.getConstructors();
            // we should check for the signature of the constructors
            // see ThematicDataDisplayLayer.getConstructorWithStringParameter
            classif = (Classification) constructors[0].newInstance(new Object[]{null, nbClasses});
            if (oldClassif != null) {
                classif.reInit(oldClassif);
            }

        } catch (Exception e) {
            System.err.println("[Colorizer] *** construction failure for " + classificationClassName);
            e.printStackTrace();

            classif = oldClassif;
        }
        // System.out.println("#Colorizer.setClassification =>"+classif);
        fireClassificationChanged();
    }

    public void resetClassification(String classificationClassName) {
        setClassification(classificationClassName, classif != null ? classif.getNumberOfClasses() : (colorScale != null && colorScale.getColorNumber() != 0) ? colorScale.getColorNumber() : 6);
    }

    public Properties getProperties(String prefix, Properties properties) {
        colorScale.getProperties(prefix, properties);
        if (classif != null) {
            properties.put(prefix + classificationClassNamePropertyString, classif.getClass().getName());
            properties.put(prefix + nbClassesPropertyString, String.valueOf(classif.getNumberOfClasses()));

            // useful in custom mode
            properties.put(prefix + BREAKS_PROPERTY_NAME, classif.getBreaksAsString());
        }

        return properties;
    }

    public void setData(DataStorageIterator dsi, String dataColumnName) {
        if (classif == null) {
            return;
        }
        Debug.message("Colorizer", "Colorizer -> CLASSIF ReInit " + classif + " * dataColName " + dataColumnName);

        classif.reInit(dsi, dataColumnName);
        fireClassificationChanged();
    }

    public Color getColor(Object value) {
        if (classif != null) {
            return colorScale.getColor(classif.getClass(value));
        }
        return Color.BLACK;
    }

    public ColorizerPanel getGUI() {
        ColorizerPanel p = new ColorizerPanel(this);
        // p.refreshPanel();
        return p;
    }

    public ColorScale getColorScale() {
        return colorScale;
    }

    public void setColorScale(ColorScale c) {
        colorScale = c;
        // Classification didn't really change, but we need to propagate
        // colorScale change.
        fireClassificationChanged();
    }

    public Classification getClassification() {
        return classif;
    }

    public String toString() {
        return "ColorScale: " + colorScale + " Classif:" + (classif == null ? "null" : classif.toString());
    }

    /** get colormap button text from the breaks get from the classification */
    public String getColorText(int i) {
        if (getClassification() != null) {
            return getClassification().getClassText(i, stringNumberFormat);
        }
        return " ";
    }

    public String[] getColorsTexts() {
        if (getClassification() == null) {
            return null;
        }
        return getClassification().getClassesTexts(stringNumberFormat);
    }

    public void setClassification(Classification classif) {
        // System.out.println("=> #Colorizer: set classif " + classif);
        if (this.classif != null && this.classif.getClass() == classif.getClass()) {
            return;
        }

        this.classif = classif;
        fireClassificationChanged();
        // System.out.println("<= #Colorizer: set classif " + this.classif);
    }
}
