// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util;

import java.awt.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.util.Vector;
import javax.swing.*;
import javax.swing.border.EmptyBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.bbn.openmap.I18n;
import com.bbn.openmap.BasicI18n;
import fr.unicaen.openmap.gui.ThemaMapColorChooser;

public class FontDialog extends JDialog implements ListSelectionListener, ActionListener,
						   WindowListener {

    /**
     * Use I18N mecanism to allow internationalization.
     */    
    public final static I18n i18n = new BasicI18n();

    /**
     * Property name for plain font style.
     */
    public final static String PLAIN_STYLE_PROPERTY_STRING = "plain";

    /**
     * Property name for bold font style.
     */
    public final static String BOLD_STYLE_PROPERTY_STRING = "bold";

    /**
     * Property name for italic font style.
     */
    public final static String ITALIC_STYLE_PROPERTY_STRING = "italic";

    /**
     * Property name for bold and italic font style.
     */
    public final static String BOLD_AND_ITALIC_STYLE_PROPERTY_STRING = "bold_italic";

    /**
     * Property name for title.
     */
    public final static String TITLE_PROPERTY_STRING = "title";

    /**
     * Property name for font.
     */
    public static final String FONT_PROPERTY_STRING = "font";

    /**
     * Property name for font style.
     */
    public static final String FONT_STYLE_PROPERTY_STRING = "font.style";

    /**
     * Property name for font size.
     */
    public static final String FONT_SIZE_PROPERTY_STRING = "font.size";

    /**
     * Property name for preview.
     */
    public final static String PREVIEW_PROPERTY_STRING = "preview";

    /**
     * Property name for previewed text.
     */
    public static final String PREVIEW_SAMPLE_PROPERTY_STRING = "preview.text";

    /**
     * Property name for "ok" button.
     */
    public static final String OK_PROPERTY_STRING = "ok";

    /**
     * Property name for "cancel" button.
     */
    public static final String CANCEL_PROPERTY_STRING = "cancel";
    
    private JTextField previewTF, fontTF, styleTF, sizeTF;
    private JButton okB, cancelB;
    private Font font;
    private String fontName = "Arial";
    private int fontStyle = 0;
    private String fontSize = "10";
    private Color fontColor = Color.BLACK;
    
    private JList fontsJL = new JList(GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames());
    private JList stylesJL = new JList(new Object [] { 
	i18n.get(FontDialog.class, PLAIN_STYLE_PROPERTY_STRING, "Plain"), 
	i18n.get(FontDialog.class, BOLD_STYLE_PROPERTY_STRING, "Bold"), 
	i18n.get(FontDialog.class, ITALIC_STYLE_PROPERTY_STRING, "Italic"),
	i18n.get(FontDialog.class, BOLD_AND_ITALIC_STYLE_PROPERTY_STRING, "Bold & Italic") }
				       );
    private JList sizesJL = new JList(new Object [] { "6", "8", "10", "12", "14", "16", "18",
						      "20", "24", "36", "48", "56", "64", "72"});

    private boolean canceled = false;


    private JPanel createList(String label, JList list, JTextField field) { //, boolean editable) {
	list.addListSelectionListener(this);
	list.setBorder(new EmptyBorder(0, 2, 0, 2));
	field.setEditable(false);
	JPanel top, middle;
	(middle = new JPanel()).setLayout(new BorderLayout());
	middle.add(field, BorderLayout.NORTH);
	middle.add(new JScrollPane(list), BorderLayout.CENTER);
	
	(top = new JPanel()).setLayout(new BorderLayout());
	top.add(middle, BorderLayout.CENTER);
	top.add(new JLabel (label), BorderLayout.NORTH);
	return top;
    }
    
    private int getFontIndex(String name) {
	for (int i = 0; i < fontsJL.getModel().getSize() ; i++)
	    if (((String) fontsJL.getModel().getElementAt(i)).equalsIgnoreCase(name)) return i;
	return -1;
    }
    
    public void selectFont(Font font) {
	int index = getFontIndex(font.getFamily());
	if (index != -1) {
	    fontsJL.setSelectedIndex(index);
	    fontsJL.ensureIndexIsVisible(index);
	}
	fontTF.setText(font.getFamily());
	
	stylesJL.setSelectedIndex(font.getStyle());
	styleTF.setText((String) stylesJL.getSelectedValue());
	sizesJL.setSelectedValue("" + font.getSize(), true);
	sizeTF.setText("" + font.getSize());
    }
    public FontDialog (Font font, boolean displayColor){
        this(font, displayColor, null);
    }

    public FontDialog (Font font, boolean displayColor, Component justificationGUI) {
	super((JFrame) null, 
	      i18n.get(FontDialog.class, 
		       TITLE_PROPERTY_STRING, 
		       "Select font ..."), 
	      true
	      );
	JPanel message = new JPanel();
	getContentPane().setLayout(new BorderLayout());
	
	this.font = font;
        previewTF = new JTextField(i18n.get(FontDialog.class,
					    PREVIEW_SAMPLE_PROPERTY_STRING,
					    "0123456789 abcdefijk ABCDEFIJK"
					    )
				   );
	previewTF.setFont(font);
		
	JPanel top, up, right, preview, bottom;
	(top = new JPanel()).setLayout(new BoxLayout(top, BoxLayout.Y_AXIS));
	(up = new JPanel()).setLayout(new BorderLayout());
	(right = new JPanel()).setLayout(new BorderLayout());
        (preview = new JPanel()).setLayout(new BorderLayout());
	bottom = new JPanel();
	
	preview.setBorder(new TitledBorder(i18n.get(FontDialog.class,
						    PREVIEW_PROPERTY_STRING,
						    "Preview")
					   )
			  );
	preview.add(previewTF, BorderLayout.CENTER);
	
        up.add(BorderLayout.CENTER, createList(i18n.get(FontDialog.class,
							FONT_PROPERTY_STRING,
							"Font"), 
					       fontsJL, 
					       fontTF = new JTextField()
					       )
	       );
        right.add(BorderLayout.CENTER, createList(i18n.get(FontDialog.class,
							   FONT_STYLE_PROPERTY_STRING,
							   "Style"),
						  stylesJL,
						  styleTF = new JTextField()
						  )
		  );
	right.add(BorderLayout.EAST, createList(i18n.get(FontDialog.class,
							 FONT_SIZE_PROPERTY_STRING,
							 "Size"),
						sizesJL,
						sizeTF = new JTextField()
						)
		  );
	sizeTF.setEditable(true);
        up.add(BorderLayout.EAST, right);
        
    JPanel colorPan = new JPanel();
    final JButton sampleColor = new JButton(" ");
    sampleColor.setBorderPainted(false);
    sampleColor.setFocusable(false);
    sampleColor.setBackground(Color.BLACK);
    sampleColor.setPreferredSize(new Dimension(50,20));
    sampleColor.addActionListener(new ActionListener(){
		public void actionPerformed(ActionEvent arg0) {
			fontColor = ThemaMapColorChooser.showDialog(sampleColor,"",fontColor);
			sampleColor.setBackground(fontColor);
		}});
    JLabel colorLabel = new JLabel("couleur");
    colorPan.add(colorLabel);
    colorPan.add(sampleColor);
        
	top.add(up);
	if(displayColor)
		top.add(colorPan);
        if(justificationGUI != null)
            top.add(justificationGUI);
	top.add(preview);
	top.setBorder(new EmptyBorder(2, 6, 0, 6));

	
	(okB = new JButton(i18n.get(FontDialog.class, OK_PROPERTY_STRING, "Ok"))).addActionListener(this);
	(cancelB = new JButton(i18n.get(FontDialog.class, CANCEL_PROPERTY_STRING, "Cancel"))).addActionListener(this);
	bottom.add(okB);
	bottom.add(cancelB);
	
        getContentPane().add(BorderLayout.CENTER, top);
	getContentPane().add(BorderLayout.SOUTH, bottom);
	
	// last, to ensure all have components have been initialized
	selectFont(font);

	addWindowListener(this);
	setSize(new Dimension(500,350));
    }
    
    public void valueChanged(ListSelectionEvent e) {
	Object source = e.getSource();

        if (source == fontsJL) {
            fontName = (String) fontsJL.getSelectedValue();
	    fontTF.setText((String) fontsJL.getSelectedValue());
        } else if ( source == stylesJL ) {
            fontStyle = stylesJL.getSelectedIndex();
	    styleTF.setText((String) stylesJL.getSelectedValue());
        } else if (source == sizesJL) {
            fontSize = (String) sizesJL.getSelectedValue();
	    sizeTF.setText((String)sizesJL.getSelectedValue());
        }
	font = new Font (fontName, fontStyle, Integer.parseInt(fontSize));
        previewTF.setFont(font);
    }
    
    public boolean isCanceled() { return canceled; }
    
    public void actionPerformed(ActionEvent e) {
	canceled = !(e.getSource() == okB);
	dispose();
    }
    
    public Font getSelectedFont() { return font; }

    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowClosing(WindowEvent e) { canceled = true; dispose(); }
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}
    
    /**
     * For testing purposes ... Some fonts will crash your
     * Java VM and you may test them directly with the
     * Chooser ...
     */
    public static void main(String s[]) {
	FontDialog fd = new FontDialog(new Font("SansSerif", 1, 14), false);
	fd.setVisible(true);
	System.exit(0);
    }

	public Color getFontColor() {
		return fontColor;
	}
}
