// <header>
/**
 *  Themamap: The Thematic Maps Production Tool
 *
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, Certic
 *  Copyright (c) 2008 Université de Caen Basse-Normandie, GREYC
 *                     		
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  <http://www.gnu.org/licenses/>
 *
 *  See http://themamap.info.unicaen.fr/authors/
 *      for a list of contributors 
 */
// </header>
package fr.unicaen.openmap.util.classification;

import java.util.ArrayList;
import java.util.Collections;
import java.lang.reflect.Constructor;
import javax.swing.*;

import fr.unicaen.openmap.util.Util;
/**
 * Perform a builtin statistical classification of data.
 * <p>You may get breaks and counts of objects per class.
 * <p>breaks[0] and breaks[breaks.length-1] are respectively equals to min and max values
 */
public abstract class Classification {
    protected int nbClasses;
    
    /**
     * @param nbClasses the number of wished classes; some methods may decide to change its value.
     */
    public Classification(int nbClasses) {
	this.nbClasses = nbClasses;
    }
    
    abstract public void reInit(fr.unicaen.openmap.dataStorage.DataStorageIterator dsi,java.lang.String nbClasses);

    public void reInit(Classification c) {
	reInit(c, c.nbClasses);
    }

    abstract public void reInit(Classification c, int nbClasses); 
    
    abstract public boolean hasNoData();

    /** used to inform the classif that its data is out of date */
    abstract public void unsetData(); 
    
    public void setNumberOfClasses(int nbClasses) {
	this.nbClasses = nbClasses;
	init();
	//	System.out.println("Classification.setNumberOfClasses("+nbClasses+") "+this);
    }

    public int getNumberOfClasses() { return nbClasses; }

    /**
     * Return a SpinnerNumberModel compatible with the classification type and the current
     * number of classes.
     */
    public SpinnerNumberModel getClassificationNumberModel() {
	return new SpinnerNumberModel (nbClasses, 1, 10, 1);
    }

    protected abstract void init();

    
    /** default implementation retruns 0. */
    public double getNthBreak(int n) { return 0; }
    protected int[] counts;
    public int[] getCounts() { return counts; }
    abstract protected void initCountsFromBreaks();    

    /**
     * return the class rank of the value d between 1 and nbClasses
     */
    
    abstract public int getClass(Object d);

    abstract public String toString();
    abstract public String dump();

// ? /    public JComponent getGUI() {
// ? /	return new ClassificationPanel(this, null);
// ? /    }

    /** returns a text describing the ith class. format is used to format the break value. */
    abstract public String getClassText(int i, String format);
    /** returns texts describing the classes. format is used to format the break value. */
    abstract public String[] getClassesTexts(String format);
    /** returns the breaks, useful for getProperties() */
    abstract public String getBreaksAsString();
    
    /** standalone tests */
    public static void main(String[] args) {
	if (args.length < 3) {
	    System.err.println("usage: java fr.unicaen.openmap.util.classification.Classification <className> <nbClasses> <nbData> [<maxValue>]");
	    return;
	}

	ArrayList<Double> data = new ArrayList<Double>();
	String className = args[0];
	int n = Integer.parseInt(args[1]);
	int N = Integer.parseInt(args[2]);
	if (args.length > 4) {
	    N = args.length - 2;
	    for (int i = 2; i < args.length; i++)
		data.add(Double.parseDouble(args[i]));
	} else {
	    double maxValue = 70;
	    if (args.length == 4)
		maxValue = Double.parseDouble(args[3]);
	    
	    for (int i = 0; i < N; i++)
		data.add((Math.random()>0.25 ? +1 : -1)*(double)maxValue*Math.random());
	}
	Classification c;
        try {
            Class cls = Class.forName("fr.unicaen.openmap.util.classification."+className);
	    Constructor[] constructors = cls.getConstructors();
	    // we should check for the signature of the constructors 
	    // see ThematicDataDisplayLayer.getConstructorWithStringParameter
	    c = (Classification) constructors[0] .newInstance(new Object [] { data, n });;
        } catch (Exception e) {
            System.err.println("construction failure for "+ className);
	    e.printStackTrace();
	    return;
        }
        
	// System.out.println(c.dump());
	//System.out.println(c);
	int nb = 0, total = 0;
	int previouscc = 1;
	    //System.out.println("Value\tClasse\tCardinal");
	for (double d : data) {
	    int cc = c.getClass(d);
	    //System.out.format("%10.4f\t%d", d, cc);
	    if (cc == previouscc)
		nb++;
	    else {
		//System.out.print("\t" + nb);
		total += nb;
		previouscc = cc;
		nb = 1;
	    }
	    //System.out.println();
	}
	total += nb;
	//System.out.println("\t\t"+ nb);
	//System.out.println("Total\t\t"+ total);
	if (total != N)
	    	//System.out.println("### "+ total + " != " + N + "!!");
	return;
    }
}
